# BEGIN BPS TAGGED BLOCK {{{
# 
# COPYRIGHT:
#  
# This software is Copyright (c) 1996-2007 Best Practical Solutions, LLC 
#                                          <jesse@bestpractical.com>
# 
# (Except where explicitly superseded by other copyright notices)
# 
# 
# LICENSE:
# 
# This work is made available to you under the terms of Version 2 of
# the GNU General Public License. A copy of that license should have
# been provided with this software, but in any event can be snarfed
# from www.gnu.org.
# 
# This work is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 or visit their web page on the internet at
# http://www.gnu.org/copyleft/gpl.html.
# 
# 
# CONTRIBUTION SUBMISSION POLICY:
# 
# (The following paragraph is not intended to limit the rights granted
# to you to modify and distribute this software under the terms of
# the GNU General Public License and is only of importance to you if
# you choose to contribute your changes and enhancements to the
# community by submitting them to Best Practical Solutions, LLC.)
# 
# By intentionally submitting any modifications, corrections or
# derivatives to this work, or any other work intended for use with
# Request Tracker, to Best Practical Solutions, LLC, you confirm that
# you are the copyright holder for those contributions and you grant
# Best Practical Solutions,  LLC a nonexclusive, worldwide, irrevocable,
# royalty-free, perpetual, license to use, copy, create derivative
# works based on those contributions, and sublicense and distribute
# those contributions and any derivatives thereof.
# 
# END BPS TAGGED BLOCK }}}
package RT::CustomField;

use strict;
no warnings qw(redefine);

sub SetMandatory {
    my $self = shift;
    my $value = shift;

    if ($value == 1) {
        my ($id, $msg) = $self->SetAttribute(Name        => 'Mandatory',
                                             Description =>'Whether the field is mandatory',
                                             Content      => $value);
    }
    if ($value == 0) {
	    my ($id, $msg) = $self->DeleteAttribute('Mandatory');
    }

    return (0, $self->IsMandatory() ? "mandatory" : "not mandatory");
}

sub IsMandatory {
    my $self = shift;
    return $self->Attributes->Named('Mandatory');
}

sub AddValueForObject {
    my $self = shift;
    my %args = (
        Object       => undef,
        Content      => undef,
        LargeContent => undef,
        ContentType  => undef,
        @_
    );
    my $obj = $args{'Object'} or return;

    unless ( $self->CurrentUserHasRight('ModifyCustomField') ) {
        return ( 0, $self->loc('Permission Denied') );
    }

    # Check Uniqueness for Asset unique CF fields
    if ($self->IsUnique()) {
      my $atype = RTx::AssetTracker::Type->new($RT::SystemUser);
      $atype->Load($args{'Object'}->Type);
      my $Assets = RTx::AssetTracker::Assets->new($RT::SystemUser);
      if (grep {$_ eq "All"} @{$self->UniqueIn()}) { # Global uniqueness
        $Assets->LimitCustomField('CUSTOMFIELD' => $self->Name,
                                  'OPERATOR'    => '=',
                                  'VALUE'       => $args{Content});
        if ($Assets->Count)  {
          my @r;
          while (my $ar = $Assets->Next()) {
            push @r, $ar->Name;
          }
          return (0, "Custom Field " . $self->Name . " is not unique across all Asset types - these Assets have the same value ($args{Content}): " . join(', ',@r));
        }
      }
      elsif (grep {$_ eq $atype->Name} @{$self->UniqueIn()}) {
        foreach my $utype (@{$self->UniqueIn()}) {
          if ($self->IsUniqueEach()) { # Unique within each selected Asset Type
            next unless $atype->Name eq $utype;
          }
          $Assets->UnLimit;
          $Assets->LimitType('OPERATOR' => '=',
                             'VALUE'    => $utype);
          $Assets->LimitCustomField('CUSTOMFIELD' => $self->Name,
                                    'OPERATOR'    => '=',
                                    'VALUE'       => $args{Content});
          if ($Assets->Count)  {
            my @r;
            while (my $ar = $Assets->Next()) {
              push @r, $ar->Name;
            }
            return (0, "Custom Field " . $self->Name . " is not unique within" . ($self->IsUniqueEach() ? " each " : " all ") . "of Assets of types \"" . $self->PrettyUnique ."\" - these Assets have the same value ($args{Content}): " . join(', ',@r));
          }
        }
      }    
    }


    # Mandatory fields cannot be null, so we do the extra check
    if ( $self->IsMandatory() ) { 
    	unless ($args{'Content'} =~ m/\S/) {
        	return ( 0, $self->loc($self->Name .' is mandatory and cannot be set to null') );
    	}
    }

    unless ( $self->MatchPattern($args{Content}) ) {
        return ( 0, $self->loc('Input for ' . $self->Name . ' must match [_1]', $self->FriendlyPattern) );
    }

    # Select fields must be validated against the list
    if ( $self->Type =~ /(?:Select)/ ) {
        my $valid_values = $self->Values;
	my $on_the_list = 0;
	while (my $value = $valid_values->Next) {
		if ($value->Name eq $args{'Content'}) {
			$on_the_list = 1;
			last;
		}
	}

	unless ($on_the_list or (!$self->IsMandatory() and !length($args{'Content'}))) { 
        	return ( 0, $self->loc('"' . $args{'Content'} . '"' . ' is not a valid value for '. $self->Name .' [_1]', $self->FriendlyPattern) );
	}
    }

    $RT::Handle->BeginTransaction;

    my $current_values = $self->ValuesForObject($obj);

    if ( $self->MaxValues ) {
        my $extra_values = ( $current_values->Count + 1 ) - $self->MaxValues;

        # (The +1 is for the new value we're adding)

        # If we have a set of current values and we've gone over the maximum
        # allowed number of values, we'll need to delete some to make room.
        # which former values are blown away is not guaranteed

        while ($extra_values) {
            my $extra_item = $current_values->Next;

            unless ( $extra_item->id ) {
                $RT::Logger->crit(
			"We were just asked to delete a custom fieldvalue that doesn't exist!"
                );
                $RT::Handle->Rollback();
                return (undef);
            }
            $extra_item->Delete;
            $extra_values--;

        }
    }
    my $newval = RT::ObjectCustomFieldValue->new( $self->CurrentUser );
    my $val    = $newval->Create(
        ObjectType   => ref($obj),
        ObjectId     => $obj->Id,
        Content      => $args{'Content'},
        LargeContent => $args{'LargeContent'},
        ContentType  => $args{'ContentType'},
        CustomField  => $self->Id
    );

    unless ($val) {
        $RT::Handle->Rollback();
        return ($val);
    }

    $RT::Handle->Commit();
    return ($val);

}

sub DeleteValueForObject {
    my $self = shift;
    my %args = ( Object => undef,
                 Content => undef,
                 Id => undef,
                     @_ );

    unless ($self->CurrentUserHasRight('ModifyCustomField')) {
        return (0, $self->loc('Permission Denied'));
    }

    my $oldval = RT::ObjectCustomFieldValue->new($self->CurrentUser);

    if (my $id = $args{'Id'}) {
        $oldval->Load($id);
    }
    unless ($oldval->id) {
        $oldval->LoadByObjectContentAndCustomField(
            Object => $args{'Object'},
            Content =>  $args{'Content'},
            CustomField => $self->Id,
        );
    }


    # check to make sure we found it
    unless ($oldval->Id) {
        return(0, $self->loc("Custom field value [_1] could not be found for custom field [_2]", $args{'Content'}, $self->Name));
    }

    # Mandatory fields cannot be null, so we do the extra check
    if ( $self->IsMandatory() ) {
    	my $current_values = $self->ValuesForObject($args{Object});
	    if ($current_values->Count == 1) {
      return ( 0, $self->loc($self->Name .' is mandatory and cannot be set to null') );
      }
    }

    # delete it

    my $ret = $oldval->Delete();
    unless ($ret) {
        return(0, $self->loc("Custom field value could not be found"));
    }
    return($oldval->Id, $self->loc("Custom field value deleted"));
}

1;
